// -----------------------------------------------------------------------
// Litmus - DKFZ SIDT unit testing library
//
// Copyright (c) German Cancer Research Center (DKFZ),
// Software development for Integrated Diagnostic and Therapy (SIDT).
// ALL RIGHTS RESERVED.
// See litmusCopyright.txt or
// http://www.dkfz.de/en/sidt/projects/litmus/copyright.html
//
// This software is distributed WITHOUT ANY WARRANTY; without even
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the above copyright notices for more information.
//
//------------------------------------------------------------------------
/*!
// @file
// @version $Revision: 4433 $ (last changed revision)
// @date    $Date: 2013-01-22 14:38:20 +0100 (Di, 22 Jan 2013) $ (last change date)
// @author  $Author: floca $ (last changed by)
*/


#ifndef __LIT_POINT_SET_TESTER_TPP
#define __LIT_POINT_SET_TESTER_TPP

#include <assert.h>

namespace lit 
{

    template<class TPointSet>
    StringType
      PointSetTester<TPointSet>::
      getTestDescription(void) const
    {
      return "Checks if actual point set equals expected point set";
    };

    template<class TPointSet>
    StringType
      PointSetTester<TPointSet>::
      getTestName(void) const
    {
      OStringStream stream;
      stream << "Check equal: actual point set == expected point set (tolerance: +-"<<_checkThreshold<<")";

      return stream.str();
    };

    template<class TPointSet>
    void
      PointSetTester<TPointSet>::
      setExpectedPointSet( const PointSetType* pPointSet)
    {
      _spExpectedPointSet = pPointSet;
    };

    template<class TPointSet>
    void
      PointSetTester<TPointSet>::
      setActualPointSet(const PointSetType* pPointSet)
    {
      _spActualPointSet = pPointSet;
    };

    template<class TPointSet>
    void
      PointSetTester<TPointSet>::
      setCheckThreshold(double checkThreshold)
    {
      _checkThreshold = checkThreshold;
    };

    template<class TPointSet>
    const typename PointSetTester<TPointSet>::PointSetType*
      PointSetTester<TPointSet>::
      getExpectedPointSet() const
    {
      return _spExpectedPointSet;
    };

    template<class TPointSet>
    const typename PointSetTester<TPointSet>::PointSetType*
      PointSetTester<TPointSet>::
      getActualPointSet() const
    {
      return _spActualPointSet;
    };


    template<class TPointSet>
    double
      PointSetTester<TPointSet>::
      getCheckThreshold() const
    {
      return _checkThreshold;
    };

    template<class TPointSet>
    PointSetTester<TPointSet>::
      PointSetTester() : _checkThreshold(0.0)
    {
    };

    template<class TPointSet>
    PointSetTester<TPointSet>::
      ~PointSetTester()
    {
    };

    template<class TPointSet>
    bool
      PointSetTester<TPointSet>::
      doCheck(void) const
    {
      bool pointSetEqualsPointSet = true;
      _maxError = itk::NumericTraits<typename PointType::RealType>::Zero;
      _errorPointCount = 0;
      _maxErrorPointExp.Fill(0);
      _maxErrorPointAct.Fill(0);

      _pResults->onTestStart(getCurrentTestLabel());


      if (_spActualPointSet->GetPoints()->Size() == _spExpectedPointSet->GetPoints()->Size())
      {
        typename PointSetType::PointsContainerConstIterator actIter = _spActualPointSet->GetPoints()->Begin();
        typename PointSetType::PointsContainerConstIterator expIter = _spExpectedPointSet->GetPoints()->Begin();

        typename PointSetType::PointsContainerConstIterator actEnd = _spActualPointSet->GetPoints()->End();
        typename PointSetType::PointsContainerConstIterator expEnd = _spExpectedPointSet->GetPoints()->End();

        while(expIter != expEnd)
        {
          typename PointType::RealType difference = (actIter.Value()).EuclideanDistanceTo(expIter.Value());

          if (difference>_checkThreshold)
          {
            pointSetEqualsPointSet = false;
            ++_errorPointCount;
            if (_maxError<difference)
            {
              _maxError = difference;
              _maxErrorPointAct = actIter.Value();
              _maxErrorPointExp = expIter.Value();
            };
          }
          ++expIter;
          ++actIter;
        }
      }
      else
      {
        pointSetEqualsPointSet = false;
      }

      return pointSetEqualsPointSet;
    };

    template<class TPointSet>
    void
      PointSetTester<TPointSet>::
      handleSuccess(void) const
    {
      OStringStream stream;
      stream << "Points of "<<_spActualPointSet << std::endl <<"equals (tolerance: "<<_checkThreshold<<")"<<std::endl<<std::endl<< _spExpectedPointSet;

      _pResults->onTestSuccess(getCurrentTestLabel(), stream.str());
    };

    template<class TPointSet>
    void
      PointSetTester<TPointSet>::
      handleFailure(void) const
    {
      OStringStream stream;
      stream << "Points of "<<_spActualPointSet << std::endl << " doesn't equal (tolerance: "<<_checkThreshold<<") "<<std::endl<<std::endl<< _spExpectedPointSet << std::endl;
      if (!(_spActualPointSet->GetPoints()->Size() == _spExpectedPointSet->GetPoints()->Size()))
      {
        stream << std::endl << "Point sets differ in size." << std::endl;
      }
      stream << std::endl << "Incorrect points count:   "<< _errorPointCount << std::endl;
      stream << std::endl << "Maximum point difference:  "<< _maxError << std::endl;
      stream << std::endl << "Maximum error points:   "<< _maxErrorPointExp << "Actual "<<_maxErrorPointAct << std::endl;
      stream << std::endl << "<DartMeasurement name=\"Incorrect points count\" type=\"numeric/int\">"<< _errorPointCount << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Maximium error (difference)\" type=\"numeric/double\">"<< _maxError << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Maximum error points\" type=\"string\">Expected "<< _maxErrorPointExp << "Actual "<<_maxErrorPointAct<<"</DartMeasurement>" << std::endl;

      _pResults->onTestFailure(getCurrentTestLabel(), stream.str());
    };

}

#endif
